# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Helpers for MJCF elements to interact with `dm_control.mujoco.Physics`."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import re
import weakref

from absl import flags
from absl import logging
from dm_control import mujoco
from dm_control.mjcf import constants
from dm_control.mjcf import debugging
from dm_control.mujoco import wrapper as mujoco_wrapper
from dm_control.mujoco.wrapper.mjbindings import sizes
import numpy as np
import six
from six.moves import range

FLAGS = flags.FLAGS
flags.DEFINE_boolean('pymjcf_log_xml', False,
                     'Whether to log the generated XML on model compilation.')

_XML_PRINT_SHARD_SIZE = 100
_PICKLING_NOT_SUPPORTED = 'Objects of type {type} cannot be pickled.'

_Attribute = collections.namedtuple(
    'Attribute',
    ('name', 'get_named_indexer', 'triggers_dirty', 'disable_on_write'))


# Object types that can be used to index into numpy arrays.
_TUPLE = tuple
_ARRAY_LIKE = (np.ndarray, list)
# We don't need to differentiate between other types of index, e.g. integers,
# slices.


def _get_index_type(index):
  for index_type in (_TUPLE, _ARRAY_LIKE):
    if isinstance(index, index_type):
      return index_type


def _pymjcf_log_xml():
  """Returns True if the generated XML should be logged on model compilation."""
  if FLAGS.is_parsed():
    return FLAGS.pymjcf_log_xml
  else:
    return FLAGS['pymjcf_log_xml'].default


def _get_attributes(size_names, strip_prefixes):
  """Creates a dict of valid attribute from Mujoco array size name."""
  strip_regex = re.compile(r'\A({})_'.format('|'.join(strip_prefixes)))
  strip = lambda string: strip_regex.sub('', string)
  out = {}
  for name, size in six.iteritems(sizes.array_sizes['mjdata']):
    if size[0] in size_names:
      attrib_name = strip(name)
      named_indexer_getter = (
          lambda physics, name=name: getattr(physics.named.data, name))
      triggers_dirty = attrib_name in constants.MJDATA_TRIGGERS_DIRTY
      out[attrib_name] = _Attribute(
          name=attrib_name,
          get_named_indexer=named_indexer_getter,
          triggers_dirty=triggers_dirty,
          disable_on_write=())
  for name, size in six.iteritems(sizes.array_sizes['mjmodel']):
    if size[0] in size_names:
      attrib_name = strip(name)
      named_indexer_getter = (
          lambda physics, name=name: getattr(physics.named.model, name))
      triggers_dirty = attrib_name not in constants.MJMODEL_DOESNT_TRIGGER_DIRTY
      disable_on_write = constants.MJMODEL_DISABLE_ON_WRITE.get(name, ())
      out[attrib_name] = _Attribute(
          name=attrib_name,
          get_named_indexer=named_indexer_getter,
          triggers_dirty=triggers_dirty,
          disable_on_write=disable_on_write)
  return out


# Fields related to the internal states of actuators (i.e. with a leading
# dimension of 'na') require special treatment.
def _get_actuator_state_fields():
  actuator_state_fields = []
  for sizes_dict in six.itervalues(sizes.array_sizes):
    for field_name, dimensions in six.iteritems(sizes_dict):
      if dimensions[0] == 'na':
        actuator_state_fields.append(field_name)
  return frozenset(actuator_state_fields)

_ACTUATOR_STATE_FIELDS = _get_actuator_state_fields()


def _filter_stateful_actuators(physics, actuator_names):
  """Removes any stateless actuators from the list of actuator names."""
  if physics.model.na:
    # MuJoCo requires that stateful actuators always come after stateless
    # actuators in the model, so we keep actuator names only if their
    # corresponding IDs are >= to the total number of stateless actuators.
    num_stateless_actuators = physics.model.nu - physics.model.na
    return [
        name for name in actuator_names
        if physics.model.name2id(name, 'actuator') >= num_stateless_actuators]
  else:
    return []


_ATTRIBUTES = {
    'actuator': _get_attributes(['na', 'nu'], strip_prefixes=['actuator']),
    'body': _get_attributes(['nbody'], strip_prefixes=['body']),
    'mocap_body': _get_attributes(['nbody', 'nmocap'], strip_prefixes=['body']),
    'camera': _get_attributes(['ncam'], strip_prefixes=['cam']),
    'equality': _get_attributes(['neq'], strip_prefixes=['eq']),
    'geom': _get_attributes(['ngeom'], strip_prefixes=['geom']),
    'hfield': _get_attributes(['nhfield'], strip_prefixes=['hfield']),
    'joint': _get_attributes(['nq', 'nv', 'njnt'],
                             strip_prefixes=['jnt', 'dof']),
    'light': _get_attributes(['nlight'], strip_prefixes=['light']),
    'material': _get_attributes(['nmat'], strip_prefixes=['mat']),
    'mesh': _get_attributes(['nmesh'], strip_prefixes=['mesh']),
    'numeric': _get_attributes(['nnumeric', 'nnumericdata'],
                               strip_prefixes=['numeric']),
    'sensor': _get_attributes(['nsensor', 'nsensordata'],
                              strip_prefixes=['sensor']),
    'site': _get_attributes(['nsite'], strip_prefixes=['site']),
    'tendon': _get_attributes(['ntendon'], strip_prefixes=['tendon', 'ten']),
    'text': _get_attributes(['ntext', 'ntextdata'], strip_prefixes=['text']),
    'texture': _get_attributes(['ntex'], strip_prefixes=['tex']),
}


def names_from_elements(mjcf_elements):
  """Returns `namespace` and `named_index` for `mjcf_elements`.

  Args:
    mjcf_elements: Either an `mjcf.Element`, or an iterable of `mjcf.Element`
        of the same kind.

  Returns:
    A tuple of `(namespace, named_indices)` where
      -`namespace` is the Mujoco element type (eg: 'geom', 'body', etc.)
      -`named_indices` are the names of `mjcf_elements`, either as a single
        string or an iterable of strings depending on whether `mjcf_elements`
        was an `mjcf.Element` or an iterable of `mjcf_Element`s.

  Raises:
      ValueError: If `mjcf_elements` cannot be bound to this Physics.
  """
  if isinstance(mjcf_elements, collections.Iterable):
    elements_tuple = tuple(mjcf_elements)
    if elements_tuple:
      namespace = _get_namespace(elements_tuple[0])
    else:
      return None, None
    for element in elements_tuple:
      element_namespace = _get_namespace(element)
      if element_namespace != namespace:
        raise ValueError('Cannot bind to a collection containing multiple '
                         'element types ({!r} != {!r}).'
                         .format(element_namespace, namespace))
    named_index = [element.full_identifier for element in elements_tuple]
  else:
    namespace = _get_namespace(mjcf_elements)
    named_index = mjcf_elements.full_identifier

  return namespace, named_index


class SynchronizingArrayWrapper(np.ndarray):
  """A non-contiguous view of an ndarray that synchronizes with the original.

  Note: this class should not be instantiated directly.
  """
  __slots__ = (
      '_backing_array',
      '_backing_index',
      '_backing_index_is_array_like',
      '_physics',
      '_triggers_dirty',
      '_disable_on_write',
  )

  def __new__(cls,
              backing_array,
              backing_index,
              physics,
              triggers_dirty,
              disable_on_write):
    obj = backing_array[backing_index].view(SynchronizingArrayWrapper)
    # pylint: disable=protected-access
    obj._backing_array = backing_array
    obj._backing_index = backing_index
    # Performance optimization: avoid repeatedly checking the type of the
    # backing index.
    backing_index_type = _get_index_type(backing_index)
    obj._backing_index_is_array_like = backing_index_type is _ARRAY_LIKE
    obj._physics = physics
    obj._triggers_dirty = triggers_dirty
    obj._disable_on_write = disable_on_write
    # pylint: enable=protected-access
    return obj

  def _synchronize_from_backing_array(self):
    if self._physics.is_dirty and not self._triggers_dirty:
      self._physics.forward()
    updated_values = self._backing_array[self._backing_index]
    # Faster than `super(...).__setitem__(slice(None), updated_values)`
    np.copyto(self, updated_values)

  def copy(self, order='C'):
    return np.copy(self, order=order)

  def __copy__(self):
    return self.copy()

  def __deepcopy__(self, memo):
    return self.copy()

  def __reduce__(self):
    raise NotImplementedError(_PICKLING_NOT_SUPPORTED.format(type=type(self)))

  def __setitem__(self, index, value):
    if self._physics.is_dirty and not self._triggers_dirty:
      self._physics.forward()
    super(SynchronizingArrayWrapper, self).__setitem__(index, value)

    # Performance optimization: avoid repeatedly checking the type of the index.
    index_type = _get_index_type(index)

    if self._backing_index_is_array_like:
      if index_type is _TUPLE:
        resolved_index = (self._backing_index[index[0]],) + index[1:]
      else:
        resolved_index = self._backing_index[index]
      self._backing_array[resolved_index] = value

    for backing_array, backing_index in self._disable_on_write:
      if index_type is _TUPLE:
        # We only need the row component of the index.
        resolved_index = backing_index[index[0]]
      elif index_type is _ARRAY_LIKE:
        resolved_index = backing_index[index]
      else:
        # If it is only an index into the columns of the backing array then we
        # just discard it and use the backing index.
        resolved_index = backing_index
      backing_array[resolved_index] = 0

    if self._triggers_dirty:
      self._physics.mark_as_dirty()

  def __setslice__(self, start, stop, value):
    self.__setitem__(slice(start, stop, None), value)


class Binding(object):
  """Binding between a mujoco.Physics and an mjcf.Element or a list of Elements.

  This object should normally be created by calling `physics.bind(element)`
  where `physics` is an instance of `mjcf.Physics`. See docstring for that
  function for details.
  """
  __slots__ = (
      '_attributes',
      '_physics',
      '_namespace',
      '_named_index',
      '_named_indexers',
      '_getattr_cache',
      '_array_index_cache',
  )

  def __init__(self, physics, namespace, named_index):
    try:
      self._attributes = _ATTRIBUTES[namespace]
    except KeyError:
      raise ValueError('elements of type {!r} cannot be bound to physics'
                       .format(namespace))
    self._physics = physics
    self._namespace = namespace
    self._named_index = named_index
    self._named_indexers = {}
    self._getattr_cache = {}
    self._array_index_cache = {}

  def __dir__(self):
    return sorted(set(dir(type(self))).union(self._attributes))

  def _get_cached_named_indexer(self, name):
    named_indexer = self._named_indexers.get(name)
    if named_indexer is None:
      try:
        named_indexer = self._attributes[name].get_named_indexer(self._physics)
        self._named_indexers[name] = named_indexer
      except KeyError:
        raise AttributeError('bound element <{}> does not have attribute {!r}'
                             .format(self._namespace, name))
    return named_indexer

  def _get_cached_array_and_index(self, name):
    """Returns `(array, index)` for a given field name."""
    named_indexer = self._get_cached_named_indexer(name)
    array = named_indexer._field  # pylint: disable=protected-access
    try:
      index = self._array_index_cache[name]
    except KeyError:
      # If we are indexing into a field relating to actuator internal states
      # then we must first remove the names of any stateless actuators.
      if name in _ACTUATOR_STATE_FIELDS:
        named_index = _filter_stateful_actuators(
            self._physics, self._named_index)
      else:
        named_index = self._named_index
      index = named_indexer._convert_key(named_index)  # pylint: disable=protected-access
      self._array_index_cache[name] = index
    return array, index

  @property
  def element_id(self):
    """The ID number of this element within MuJoCo's data structures."""
    try:
      element_id = self._getattr_cache['element_id']
    except KeyError:
      if isinstance(self._named_index, list):
        element_id = np.array(
            [self._physics.model.name2id(item_name, self._namespace)
             for item_name in self._named_index])
        element_id.flags.writeable = False
      else:
        element_id = self._physics.model.name2id(
            self._named_index, self._namespace)
      self._getattr_cache['element_id'] = element_id
    return element_id

  def __getattr__(self, name):
    if name in Binding.__slots__:
      return super(Binding, self).__getattr__(name)
    else:
      try:
        out = self._getattr_cache[name]
        out._synchronize_from_backing_array()  # pylint: disable=protected-access
      except KeyError:
        array, index = self._get_cached_array_and_index(name)
        triggers_dirty = self._attributes[name].triggers_dirty

        # A list of (array, index) tuples specifying other addresses that need
        # to be zeroed out when this array attribute is written to.
        disable_on_write = []
        for name_to_disable in self._attributes[name].disable_on_write:
          array_to_disable, index_to_disable = self._get_cached_array_and_index(
              name_to_disable)
          # Ensure that the result of indexing is a `SynchronizingArrayWrapper`
          # rather than a scalar, otherwise we won't be able to write into it.
          if array_to_disable.ndim == 1:
            if isinstance(index_to_disable, np.ndarray):
              index_to_disable = index_to_disable.copy().reshape(-1, 1)
            else:
              index_to_disable = [index_to_disable]
          disable_on_write.append((array_to_disable, index_to_disable))

        if self._physics.is_dirty and not triggers_dirty:
          self._physics.forward()
        if isinstance(index, int) and array.ndim == 1:
          # Case where indexing results in a scalar.
          out = array[index]
        else:
          # Case where indexing results in an array.
          out = SynchronizingArrayWrapper(
              backing_array=array,
              backing_index=index,
              physics=self._physics,
              triggers_dirty=triggers_dirty,
              disable_on_write=disable_on_write)
          self._getattr_cache[name] = out
      return out

  def __setattr__(self, name, value):
    if name in Binding.__slots__:
      super(Binding, self).__setattr__(name, value)
    else:
      if self._physics.is_dirty and not self._attributes[name].triggers_dirty:
        self._physics.forward()
      array, index = self._get_cached_array_and_index(name)
      array[index] = value
      for name_to_disable in self._attributes[name].disable_on_write:
        disable_array, disable_index = self._get_cached_array_and_index(
            name_to_disable)
        disable_array[disable_index] = 0
      if self._attributes[name].triggers_dirty:
        self._physics.mark_as_dirty()

  def _get_name_and_indexer_and_expression(self, index):
    """Returns (name, indexer, expression) for a given input to __getitem__."""
    if isinstance(index, tuple):
      name, column_index = index
      try:
        # If named_index and column_index are both array-like, reshape
        # named_index to (n, 1) so that it can be broadcasted against
        # column_index.
        expression = np.ix_(self._named_index, column_index)
      except ValueError:
        expression = (self._named_index, column_index)
    else:
      name = index
      expression = self._named_index
    indexer = self._get_cached_named_indexer(name)
    return name, indexer, expression

  def __getitem__(self, index):
    name, indexer, expression = self._get_name_and_indexer_and_expression(index)
    if self._physics.is_dirty and not self._attributes[name].triggers_dirty:
      self._physics.forward()
    return indexer[expression]

  def __setitem__(self, index, value):
    name, indexer, expression = self._get_name_and_indexer_and_expression(index)
    if self._physics.is_dirty and not self._attributes[name].triggers_dirty:
      self._physics.forward()
    indexer[expression] = value
    if self._attributes[name].triggers_dirty:
      self._physics.mark_as_dirty()


class _EmptyBinding(object):
  """The result of binding no `mjcf.Elements` to an `mjcf.Physics` instance."""

  __slots__ = ('_arr',)

  def __init__(self):
    self._arr = np.empty((0))

  def __getattr__(self, name):
    return self._arr

  def __setattr__(self, name, value):
    if name in self.__slots__:
      super(_EmptyBinding, self).__setattr__(name, value)
    else:
      raise ValueError('Cannot assign a value to an empty binding.')

_EMPTY_BINDING = _EmptyBinding()


def _log_xml(xml_string):
  xml_lines = xml_string.split('\n')
  for start_line in range(0, len(xml_lines), _XML_PRINT_SHARD_SIZE):
    end_line = min(start_line + _XML_PRINT_SHARD_SIZE, len(xml_lines))
    template = 'XML lines %d-%d of %d:\n%s'
    if start_line == 0:
      template = 'PyMJCF: compiling generated XML:\n' + template
    logging.info(template, start_line + 1, end_line, len(xml_lines),
                 '\n'.join(xml_lines[start_line:end_line]))


class Physics(mujoco.Physics):
  """A specialized `mujoco.Physics` that supports binding to MJCF elements."""

  @classmethod
  def from_mjcf_model(cls, mjcf_model):
    """Constructs a new `mjcf.Physics` from an `mjcf.RootElement`.

    Args:
      mjcf_model: An `mjcf.RootElement` instance.

    Returns:
      A new `mjcf.Physics` instance.
    """
    debug_context = debugging.DebugContext()
    xml_string = mjcf_model.to_xml_string(debug_context=debug_context)
    if _pymjcf_log_xml():
      if debug_context.debug_mode and debug_context.default_dump_dir:
        logging.info('Full debug info is dumped to disk at %s',
                     debug_context.default_dump_dir)
        debug_context.dump_full_debug_info_to_disk()
      else:
        logging.info('Full debug info is not yet dumped to disk. If this is '
                     'needed, pass all three flags: --pymjcf_log_xml '
                     '--pymjcf_debug --pymjcf_debug_full_dump_dir=/path/dir/')
      _log_xml(xml_string)
    assets = mjcf_model.get_assets()
    try:
      return cls.from_xml_string(xml_string=xml_string, assets=assets)
    except mujoco_wrapper.Error:
      debug_context.process_and_raise_last_exception()

  def reload_from_mjcf_model(self, mjcf_model):
    """Reloads this `mjcf.Physics` from an `mjcf.RootElement`.

    After calling this method, the state of this `Physics` instance is the same
    as a new `Physics` instance created with the `from_mjcf_model` named
    constructor.

    Args:
      mjcf_model: An `mjcf.RootElement` instance.
    """
    debug_context = debugging.DebugContext()
    xml_string = mjcf_model.to_xml_string(debug_context=debug_context)
    if _pymjcf_log_xml():
      _log_xml(xml_string)
    assets = mjcf_model.get_assets()
    try:
      self.reload_from_xml_string(xml_string=xml_string, assets=assets)
    except mujoco_wrapper.Error:
      debug_context.process_and_raise_last_exception()

  def _reload_from_data(self, data):
    """Initializes a new or existing `Physics` instance from a `core.MjData`.

    Assigns all attributes and sets up rendering contexts and named indexing.

    The default constructor as well as the other `reload_from` methods should
    delegate to this method.

    Args:
      data: Instance of `core.MjData`.
    """
    super(Physics, self)._reload_from_data(data)
    self._bindings = {}
    self._bindings[()] = _EMPTY_BINDING
    self._dirty = False

  @property
  def is_dirty(self):
    """Whether this physics' internal state needs to be recalculated."""
    return self._dirty

  def mark_as_dirty(self):
    """Marks this physics as dirty, thus requiring recalculation."""
    self._dirty = True

  def forward(self):
    """Recomputes the forward dynamics without advancing the simulation."""
    super(Physics, self).forward()
    self._dirty = False

  def bind(self, mjcf_elements):
    """Creates a binding between this `Physics` instance and `mjcf.Element`s.

    The binding allows for easier interaction with the `Physics` data structures
    related to an MJCF element. For example, in order to access the Cartesian
    position of a geom, we can use:

    ```python
    physics.bind(geom_element).pos
    ```

    instead of the more cumbersome:

    ```python
    physics.named.model.geom_pos[geom_element.full_identifier]
    ```

    Note that the binding takes into account the type of element. This allows us
    to remove prefixes from certain common attributes in order to unify access.
    For example, we can use:

    ```python
    physics.bind(geom_element).pos = [1, 2, 3]
    physics.bind(site_element).pos = [4, 5, 6]
    ```

    instead of:

    ```python
    physics.named.model.geom_pos[geom_element.full_identifier] = [1, 2, 3]
    physics.named.model.site_pos[site_element.full_identifier] = [4, 5, 6]
    ```

    This in turn allows for the creation of common algorithms that can operate
    across a wide range of element type.

    When attribute values are modified through the binding, future queries of
    derived values are automatically recalculated if necessary. For example,
    if a joint's `qpos` is modified and a site's `xpos` is later read, the value
    of the `xpos` is updated according to the new joint configuration. This is
    done lazily when an updated value is required, so repeated value
    modifications do not incur a performance penalty.

    It is also possible to bind a sequence containing one or more elements,
    provided they are all of the same type. In this case the binding exposes
    `SynchronizingArrayWrapper`s, which are array-like objects that provide
    writeable views onto the corresponding memory addresses in MuJoCo. Writing
    into a `SynchronizingArrayWrapper` causes the underlying values in MuJoCo
    to be updated, and if necessary causes derived values to be recalculated.
    Note that in order to trigger recalculation it is necessary to reference a
    derived attribute of a binding.

    ```python
    bound_joints = physics.bind([joint1, joint2])
    bound_bodies = physics.bind([body1, body2])
    # `qpos_view` and `xpos_view` are `SynchronizingArrayWrapper`s providing
    # views onto `physics.data.qpos` and `physics.data.xpos` respectively.
    qpos_view = bound_joints.qpos
    xpos_view = bound_bodies.xpos
    # This updates the corresponding values in `physics.data.qpos`, and marks
    # derived values (such as `physics.data.xpos`) as needing recalculation.
    qpos_view[0] += 1.
    # Note: at this point `xpos_view` still contains the old values, since we
    # need to actually read the value of a derived attribute in order to
    # trigger recalculation.
    another_xpos_view = bound_bodies.xpos  # Triggers recalculation of `xpos`.
    # Now both `xpos_view` and `another_xpos_view` will contain the updated
    # values.
    ```

    Note that `SynchronizingArrayWrapper`s cannot be pickled. We also do not
    recommend holding references to them - instead hold a reference to the
    binding object, or call `physics.bind` again.

    Bindings also support numpy-style square bracket indexing. The first element
    in the indexing expression should be an attribute name, and the second
    element (if present) is used to index into the columns of the underlying
    array. Named indexing into columns is also allowed, provided that the
    corresponding field in `physics.named` supports it.

    ```python
    physics.bind([geom1, geom2])['pos'] = [[1, 2, 3], [4, 5, 6]]
    physics.bind([geom1, geom2])['pos', ['x', 'z']] = [[1, 3], [4, 6]]
    ```

    Args:
      mjcf_elements: Either an `mjcf.Element`, or an iterable of `mjcf.Element`
        of the same kind.

    Returns:
      A binding between this Physics instance an `mjcf_elements`, as described
      above.

    Raises:
      ValueError: If `mjcf_elements` cannot be bound to this Physics.
    """
    if mjcf_elements is None:
      return None

    # To reduce overhead from processing MJCF elements and making new bindings,
    # we cache and reuse existing Binding objects. The cheapest version of
    # caching is when we can use `mjcf_elements` as key directly. However, this
    # is not always possible since `mjcf_elements` may contain weak references.
    # In this case, we fallback to using the elements' namespace and identifiers
    # as cache keys instead.

    # Checking for iterability in this way is cheaper than using `isinstance`.
    try:
      cache_key = tuple(mjcf_elements)
    except TypeError:
      # `mjcf_elements` is not iterable.
      cache_key = mjcf_elements

    needs_new_binding = False
    try:
      binding = self._bindings[cache_key]
    except KeyError:
      # This means `mjcf_elements` is hashable, so we use it as cache key.
      namespace, named_index = names_from_elements(mjcf_elements)
      needs_new_binding = True
    except TypeError:
      # This means `mjcf_elements` is unhashable, fallback to caching by name.
      namespace, named_index = names_from_elements(mjcf_elements)

      # Checking for iterability in this way is cheaper than using `isinstance`.
      try:
        cache_key = (namespace, tuple(named_index))
      except TypeError:
        # `named_index` is not iterable.
        cache_key = (namespace, named_index)

      try:
        binding = self._bindings[cache_key]
      except KeyError:
        needs_new_binding = True

    if needs_new_binding:
      binding = Binding(weakref.proxy(self), namespace, named_index)
      self._bindings[cache_key] = binding

    return binding


def _get_namespace(element):
  """Returns the element namespace string."""
  # The worldbody is treated as a member of the `body` namespace.
  if element.tag == 'worldbody':
    namespace = 'body'
  else:
    namespace = element.spec.namespace.split(constants.NAMESPACE_SEPARATOR)[0]
    # Mocap bodies have distinct attributes so we use a dummy namespace for
    # them.
    if namespace == 'body' and element.mocap == 'true':
      namespace = 'mocap_body'
  return namespace



# Copyright 2019 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Tests for dm_control.locomotion.soccer.pitch."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import absltest
from absl.testing import parameterized
from dm_control import composer
from dm_control.composer.variation import distributions
from dm_control.entities import props
from dm_control.locomotion.soccer import pitch as pitch_lib
from dm_control.locomotion.soccer import team as team_lib
import numpy as np


class PitchTest(parameterized.TestCase):

  def _pitch_with_ball(self, pitch_size, ball_pos):
    pitch = pitch_lib.Pitch(size=pitch_size)
    self.assertEqual(pitch.size, pitch_size)

    sphere = props.Primitive(geom_type='sphere', size=(0.1,), pos=ball_pos)
    pitch.register_ball(sphere)
    pitch.attach(sphere)

    env = composer.Environment(
        composer.NullTask(pitch), random_state=np.random.RandomState(42))
    env.reset()
    return pitch

  def test_pitch_none_detected(self):
    pitch = self._pitch_with_ball((12, 9), (0, 0, 0))
    self.assertEmpty(pitch.detected_off_court())
    self.assertIsNone(pitch.detected_goal())

  def test_pitch_detected_off_court(self):
    pitch = self._pitch_with_ball((12, 9), (20, 0, 0))
    self.assertLen(pitch.detected_off_court(), 1)
    self.assertIsNone(pitch.detected_goal())

  def test_pitch_detected_away_goal(self):
    pitch = self._pitch_with_ball((12, 9), (-9.5, 0, 1))
    self.assertLen(pitch.detected_off_court(), 1)
    self.assertEqual(team_lib.Team.AWAY, pitch.detected_goal())

  def test_pitch_detected_home_goal(self):
    pitch = self._pitch_with_ball((12, 9), (9.5, 0, 1))
    self.assertLen(pitch.detected_off_court(), 1)
    self.assertEqual(team_lib.Team.HOME, pitch.detected_goal())

  @parameterized.parameters((True, distributions.Uniform()),
                            (False, distributions.Uniform()))
  def test_randomize_pitch(self, keep_aspect_ratio, randomizer):
    pitch = pitch_lib.RandomizedPitch(
        min_size=(4, 3),
        max_size=(8, 6),
        randomizer=randomizer,
        keep_aspect_ratio=keep_aspect_ratio)
    pitch.initialize_episode_mjcf(np.random.RandomState(42))

    self.assertBetween(pitch.size[0], 4, 8)
    self.assertBetween(pitch.size[1], 3, 6)

    if keep_aspect_ratio:
      self.assertAlmostEqual((pitch.size[0] - 4) / (8. - 4.),
                             (pitch.size[1] - 3) / (6. - 3.))


if __name__ == '__main__':
  absltest.main()



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""A dm_env.Environment subclass for control-specific environments."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc
import collections
import contextlib

import dm_env
from dm_env import specs
import numpy as np
import six
from six.moves import range

FLAT_OBSERVATION_KEY = 'observations'


class Environment(dm_env.Environment):
  """Class for physics-based reinforcement learning environments."""

  def __init__(self,
               physics,
               task,
               time_limit=float('inf'),
               control_timestep=None,
               n_sub_steps=None,
               flat_observation=False):
    """Initializes a new `Environment`.

    Args:
      physics: Instance of `Physics`.
      task: Instance of `Task`.
      time_limit: Optional `int`, maximum time for each episode in seconds. By
        default this is set to infinite.
      control_timestep: Optional control time-step, in seconds.
      n_sub_steps: Optional number of physical time-steps in one control
        time-step, aka "action repeats". Can only be supplied if
        `control_timestep` is not specified.
      flat_observation: If True, observations will be flattened and concatenated
        into a single numpy array.

    Raises:
      ValueError: If both `n_sub_steps` and `control_timestep` are supplied.
    """
    self._task = task
    self._physics = physics
    self._flat_observation = flat_observation

    if n_sub_steps is not None and control_timestep is not None:
      raise ValueError('Both n_sub_steps and control_timestep were supplied.')
    elif n_sub_steps is not None:
      self._n_sub_steps = n_sub_steps
    elif control_timestep is not None:
      self._n_sub_steps = compute_n_steps(control_timestep,
                                          self._physics.timestep())
    else:
      self._n_sub_steps = 1

    if time_limit == float('inf'):
      self._step_limit = float('inf')
    else:
      self._step_limit = time_limit / (
          self._physics.timestep() * self._n_sub_steps)
    self._step_count = 0
    self._reset_next_step = True

  def reset(self):
    """Starts a new episode and returns the first `TimeStep`."""
    self._reset_next_step = False
    self._step_count = 0
    with self._physics.reset_context():
      self._task.initialize_episode(self._physics)

    observation = self._task.get_observation(self._physics)
    if self._flat_observation:
      observation = flatten_observation(observation)

    return dm_env.TimeStep(
        step_type=dm_env.StepType.FIRST,
        reward=None,
        discount=None,
        observation=observation)

  def step(self, action):
    """Updates the environment using the action and returns a `TimeStep`."""

    if self._reset_next_step:
      return self.reset()

    self._task.before_step(action, self._physics)
    for _ in range(self._n_sub_steps):
      self._physics.step()
    self._task.after_step(self._physics)

    reward = self._task.get_reward(self._physics)
    observation = self._task.get_observation(self._physics)
    if self._flat_observation:
      observation = flatten_observation(observation)

    self._step_count += 1
    if self._step_count >= self._step_limit:
      discount = 1.0
    else:
      discount = self._task.get_termination(self._physics)

    episode_over = discount is not None

    if episode_over:
      self._reset_next_step = True
      return dm_env.TimeStep(
          dm_env.StepType.LAST, reward, discount, observation)
    else:
      return dm_env.TimeStep(dm_env.StepType.MID, reward, 1.0, observation)

  def action_spec(self):
    """Returns the action specification for this environment."""
    return self._task.action_spec(self._physics)

  def step_spec(self):
    """May return a specification for the values returned by `step`."""
    return self._task.step_spec(self._physics)

  def observation_spec(self):
    """Returns the observation specification for this environment.

    Infers the spec from the observation, unless the Task implements the
    `observation_spec` method.

    Returns:
      An dict mapping observation name to `ArraySpec` containing observation
      shape and dtype.
    """
    try:
      return self._task.observation_spec(self._physics)
    except NotImplementedError:
      observation = self._task.get_observation(self._physics)
      if self._flat_observation:
        observation = flatten_observation(observation)
      return _spec_from_observation(observation)

  @property
  def physics(self):
    return self._physics

  @property
  def task(self):
    return self._task

  def control_timestep(self):
    """Returns the interval between agent actions in seconds."""
    return self.physics.timestep() * self._n_sub_steps


def compute_n_steps(control_timestep, physics_timestep, tolerance=1e-8):
  """Returns the number of physics timesteps in a single control timestep.

  Args:
    control_timestep: Control time-step, should be an integer multiple of the
      physics timestep.
    physics_timestep: The time-step of the physics simulation.
    tolerance: Optional tolerance value for checking if `physics_timestep`
      divides `control_timestep`.

  Returns:
    The number of physics timesteps in a single control timestep.

  Raises:
    ValueError: If `control_timestep` is smaller than `physics_timestep` or if
      `control_timestep` is not an integer multiple of `physics_timestep`.
  """
  if control_timestep < physics_timestep:
    raise ValueError(
        'Control timestep ({}) cannot be smaller than physics timestep ({}).'.
        format(control_timestep, physics_timestep))
  if abs((control_timestep / physics_timestep - round(
      control_timestep / physics_timestep))) > tolerance:
    raise ValueError(
        'Control timestep ({}) must be an integer multiple of physics timestep '
        '({})'.format(control_timestep, physics_timestep))
  return int(round(control_timestep / physics_timestep))


def _spec_from_observation(observation):
  result = collections.OrderedDict()
  for key, value in six.iteritems(observation):
    result[key] = specs.Array(value.shape, value.dtype, name=key)
  return result

# Base class definitions for objects supplied to Environment.


@six.add_metaclass(abc.ABCMeta)
class Physics(object):
  """Simulates a physical environment."""

  @abc.abstractmethod
  def step(self, n_sub_steps=1):
    """Updates the simulation state.

    Args:
      n_sub_steps: Optional number of times to repeatedly update the simulation
        state. Defaults to 1.
    """

  @abc.abstractmethod
  def time(self):
    """Returns the elapsed simulation time in seconds."""

  @abc.abstractmethod
  def timestep(self):
    """Returns the simulation timestep."""

  def set_control(self, control):
    """Sets the control signal for the actuators."""
    raise NotImplementedError('set_control is not supported.')

  @contextlib.contextmanager
  def reset_context(self):
    """Context manager for resetting the simulation state.

    Sets the internal simulation to a default state when entering the block.

    ```python
    with physics.reset_context():
      # Set joint and object positions.

    physics.step()
    ```

    Yields:
      The `Physics` instance.
    """
    try:
      self.reset()
    except PhysicsError:
      pass
    yield self
    self.after_reset()

  @abc.abstractmethod
  def reset(self):
    """Resets internal variables of the physics simulation."""

  @abc.abstractmethod
  def after_reset(self):
    """Runs after resetting internal variables of the physics simulation."""

  def check_divergence(self):
    """Raises a `PhysicsError` if the simulation state is divergent.

    The default implementation is a no-op.
    """


class PhysicsError(RuntimeError):
  """Raised if the state of the physics simulation becomes divergent."""


@six.add_metaclass(abc.ABCMeta)
class Task(object):
  """Defines a task in a `control.Environment`."""

  @abc.abstractmethod
  def initialize_episode(self, physics):
    """Sets the state of the environment at the start of each episode.

    Called by `control.Environment` at the start of each episode *within*
    `physics.reset_context()` (see the documentation for `base.Physics`).

    Args:
      physics: Instance of `Physics`.
    """

  @abc.abstractmethod
  def before_step(self, action, physics):
    """Updates the task from the provided action.

    Called by `control.Environment` before stepping the physics engine.

    Args:
      action: numpy array or array-like action values, or a nested structure of
        such arrays. Should conform to the specification returned by
        `self.action_spec(physics)`.
      physics: Instance of `Physics`.
    """

  def after_step(self, physics):
    """Optional method to update the task after the physics engine has stepped.

    Called by `control.Environment` after stepping the physics engine and before
    `control.Environment` calls `get_observation, `get_reward` and
    `get_termination`.

    The default implementation is a no-op.

    Args:
      physics: Instance of `Physics`.
    """

  @abc.abstractmethod
  def action_spec(self, physics):
    """Returns a specification describing the valid actions for this task.

    Args:
      physics: Instance of `Physics`.

    Returns:
      A `BoundedArraySpec`, or a nested structure containing `BoundedArraySpec`s
      that describe the shapes, dtypes and elementwise lower and upper bounds
      for the action array(s) passed to `self.step`.
    """

  def step_spec(self, physics):
    """Returns a specification describing the time_step for this task.

    Args:
      physics: Instance of `Physics`.

    Returns:
      A `BoundedArraySpec`, or a nested structure containing `BoundedArraySpec`s
      that describe the shapes, dtypes and elementwise lower and upper bounds
      for the array(s) returned by `self.step`.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def get_observation(self, physics):
    """Returns an observation from the environment.

    Args:
      physics: Instance of `Physics`.
    """

  @abc.abstractmethod
  def get_reward(self, physics):
    """Returns a reward from the environment.

    Args:
      physics: Instance of `Physics`.
    """

  def get_termination(self, physics):
    """If the episode should end, returns a final discount, otherwise None."""

  def observation_spec(self, physics):
    """Optional method that returns the observation spec.

    If not implemented, the Environment infers the spec from the observation.

    Args:
      physics: Instance of `Physics`.

    Returns:
      A dict mapping observation name to `ArraySpec` containing observation
      shape and dtype.
    """
    raise NotImplementedError()


def flatten_observation(observation, output_key=FLAT_OBSERVATION_KEY):
  """Flattens multiple observation arrays into a single numpy array.

  Args:
    observation: A mutable mapping from observation names to numpy arrays.
    output_key: The key for the flattened observation array in the output.

  Returns:
    A mutable mapping of the same type as `observation`. This will contain a
    single key-value pair consisting of `output_key` and the flattened
    and concatenated observation array.

  Raises:
    ValueError: If `observation` is not a `collections.MutableMapping`.
  """
  if not isinstance(observation, collections.MutableMapping):
    raise ValueError('Can only flatten dict-like observations.')

  if isinstance(observation, collections.OrderedDict):
    keys = six.iterkeys(observation)
  else:
    # Keep a consistent ordering for other mappings.
    keys = sorted(six.iterkeys(observation))

  observation_arrays = [observation[key].ravel() for key in keys]
  return type(observation)([(output_key, np.concatenate(observation_arrays))])



# Copyright 2020 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Bowl arena with bumps."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from dm_control import composer
from dm_control.locomotion.arenas import assets as locomotion_arenas_assets
from dm_control.mujoco.wrapper import mjbindings

import numpy as np
from scipy import ndimage

mjlib = mjbindings.mjlib

_TOP_CAMERA_DISTANCE = 100
_TOP_CAMERA_Y_PADDING_FACTOR = 1.1

# Constants related to terrain generation.
_TERRAIN_SMOOTHNESS = .5  # 0.0: maximally bumpy; 1.0: completely smooth.
_TERRAIN_BUMP_SCALE = .2  # Spatial scale of terrain bumps (in meters).


class Bowl(composer.Arena):
  """A bowl arena with sinusoidal bumps."""

  def _build(self, size=(10, 10), aesthetic='default', name='bowl'):
    super(Bowl, self)._build(name=name)

    self._hfield = self._mjcf_root.asset.add(
        'hfield',
        name='terrain',
        nrow=201,
        ncol=201,
        size=(6, 6, 0.5, 0.1))

    if aesthetic != 'default':
      ground_info = locomotion_arenas_assets.get_ground_texture_info(aesthetic)
      sky_info = locomotion_arenas_assets.get_sky_texture_info(aesthetic)
      texturedir = locomotion_arenas_assets.get_texturedir(aesthetic)
      self._mjcf_root.compiler.texturedir = texturedir

      self._texture = self._mjcf_root.asset.add(
          'texture', name='aesthetic_texture', file=ground_info.file,
          type=ground_info.type)
      self._material = self._mjcf_root.asset.add(
          'material', name='aesthetic_material', texture=self._texture,
          texuniform='true')
      self._skybox = self._mjcf_root.asset.add(
          'texture', name='aesthetic_skybox', file=sky_info.file,
          type='skybox', gridsize=sky_info.gridsize,
          gridlayout=sky_info.gridlayout)
      self._terrain_geom = self._mjcf_root.worldbody.add(
          'geom',
          name='terrain',
          type='hfield',
          pos=(0, 0, -0.01),
          hfield='terrain',
          material=self._material)
      self._ground_geom = self._mjcf_root.worldbody.add(
          'geom',
          type='plane',
          name='groundplane',
          size=list(size) + [0.5],
          material=self._material)
    else:
      self._terrain_geom = self._mjcf_root.worldbody.add(
          'geom',
          name='terrain',
          type='hfield',
          rgba=(0.2, 0.3, 0.4, 1),
          pos=(0, 0, -0.01),
          hfield='terrain')
      self._ground_geom = self._mjcf_root.worldbody.add(
          'geom',
          type='plane',
          name='groundplane',
          rgba=(0.2, 0.3, 0.4, 1),
          size=list(size) + [0.5])

    self._mjcf_root.visual.headlight.set_attributes(
        ambient=[.4, .4, .4], diffuse=[.8, .8, .8], specular=[.1, .1, .1])

    self._regenerate = True

  def regenerate(self, random_state):
    # regeneration of the bowl requires physics, so postponed to initialization.
    self._regenerate = True

  def initialize_episode(self, physics, random_state):
    if self._regenerate:
      self._regenerate = False

      # Get heightfield resolution, assert that it is square.
      res = physics.bind(self._hfield).nrow
      assert res == physics.bind(self._hfield).ncol

      # Sinusoidal bowl shape.
      row_grid, col_grid = np.ogrid[-1:1:res*1j, -1:1:res*1j]
      radius = np.clip(np.sqrt(col_grid**2 + row_grid**2), .1, 1)
      bowl_shape = .5 - np.cos(2*np.pi*radius)/2

      # Random smooth bumps.
      terrain_size = 2 * physics.bind(self._hfield).size[0]
      bump_res = int(terrain_size / _TERRAIN_BUMP_SCALE)
      bumps = random_state.uniform(_TERRAIN_SMOOTHNESS, 1, (bump_res, bump_res))
      smooth_bumps = ndimage.zoom(bumps, res / float(bump_res))

      # Terrain is elementwise product.
      terrain = bowl_shape * smooth_bumps
      start_idx = physics.bind(self._hfield).adr
      physics.model.hfield_data[start_idx:start_idx+res**2] = terrain.ravel()

      # If we have a rendering context, we need to re-upload the modified
      # heightfield data.
      if physics.contexts:
        with physics.contexts.gl.make_current() as ctx:
          ctx.call(mjlib.mjr_uploadHField,
                   physics.model.ptr,
                   physics.contexts.mujoco.ptr,
                   physics.bind(self._hfield).element_id)

  @property
  def ground_geoms(self):
    return (self._terrain_geom, self._ground_geom)



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Hopper domain."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from dm_control import mujoco
from dm_control.rl import control
from dm_control.suite import base
from dm_control.suite import common
from dm_control.suite.utils import randomizers
from dm_control.utils import containers
from dm_control.utils import rewards
import numpy as np


SUITE = containers.TaggedTasks()

_CONTROL_TIMESTEP = .02  # (Seconds)

# Default duration of an episode, in seconds.
_DEFAULT_TIME_LIMIT = 20

# Minimal height of torso over foot above which stand reward is 1.
_STAND_HEIGHT = 0.6

# Hopping speed above which hop reward is 1.
_HOP_SPEED = 2


def get_model_and_assets():
  """Returns a tuple containing the model XML string and a dict of assets."""
  return common.read_model('hopper.xml'), common.ASSETS


@SUITE.add('benchmarking')
def stand(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
  """Returns a Hopper that strives to stand upright, balancing its pose."""
  physics = Physics.from_xml_string(*get_model_and_assets())
  task = Hopper(hopping=False, random=random)
  environment_kwargs = environment_kwargs or {}
  return control.Environment(
      physics, task, time_limit=time_limit, control_timestep=_CONTROL_TIMESTEP,
      **environment_kwargs)


@SUITE.add('benchmarking')
def hop(time_limit=_DEFAULT_TIME_LIMIT, random=None, environment_kwargs=None):
  """Returns a Hopper that strives to hop forward."""
  physics = Physics.from_xml_string(*get_model_and_assets())
  task = Hopper(hopping=True, random=random)
  environment_kwargs = environment_kwargs or {}
  return control.Environment(
      physics, task, time_limit=time_limit, control_timestep=_CONTROL_TIMESTEP,
      **environment_kwargs)


class Physics(mujoco.Physics):
  """Physics simulation with additional features for the Hopper domain."""

  def height(self):
    """Returns height of torso with respect to foot."""
    return (self.named.data.xipos['torso', 'z'] -
            self.named.data.xipos['foot', 'z'])

  def speed(self):
    """Returns horizontal speed of the Hopper."""
    return self.named.data.sensordata['torso_subtreelinvel'][0]

  def touch(self):
    """Returns the signals from two foot touch sensors."""
    return np.log1p(self.named.data.sensordata[['touch_toe', 'touch_heel']])


class Hopper(base.Task):
  """A Hopper's `Task` to train a standing and a jumping Hopper."""

  def __init__(self, hopping, random=None):
    """Initialize an instance of `Hopper`.

    Args:
      hopping: Boolean, if True the task is to hop forwards, otherwise it is to
        balance upright.
      random: Optional, either a `numpy.random.RandomState` instance, an
        integer seed for creating a new `RandomState`, or None to select a seed
        automatically (default).
    """
    self._hopping = hopping
    super(Hopper, self).__init__(random=random)

  def initialize_episode(self, physics):
    """Sets the state of the environment at the start of each episode."""
    randomizers.randomize_limited_and_rotational_joints(physics, self.random)
    self._timeout_progress = 0
    super(Hopper, self).initialize_episode(physics)

  def get_observation(self, physics):
    """Returns an observation of positions, velocities and touch sensors."""
    obs = collections.OrderedDict()
    # Ignores horizontal position to maintain translational invariance:
    obs['position'] = physics.data.qpos[1:].copy()
    obs['velocity'] = physics.velocity()
    obs['touch'] = physics.touch()
    return obs

  def get_reward(self, physics):
    """Returns a reward applicable to the performed task."""
    standing = rewards.tolerance(physics.height(), (_STAND_HEIGHT, 2))
    if self._hopping:
      hopping = rewards.tolerance(physics.speed(),
                                  bounds=(_HOP_SPEED, float('inf')),
                                  margin=_HOP_SPEED/2,
                                  value_at_margin=0.5,
                                  sigmoid='linear')
      return standing * hopping
    else:
      small_control = rewards.tolerance(physics.control(),
                                        margin=1, value_at_margin=0,
                                        sigmoid='quadratic').mean()
      small_control = (small_control + 4) / 5
      return standing * small_control



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Base class for tasks in the Control Suite."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from dm_control import mujoco
from dm_control.rl import control

import numpy as np


class Task(control.Task):
  """Base class for tasks in the Control Suite.

  Actions are mapped directly to the states of MuJoCo actuators: each element of
  the action array is used to set the control input for a single actuator. The
  ordering of the actuators is the same as in the corresponding MJCF XML file.

  Attributes:
    random: A `numpy.random.RandomState` instance. This should be used to
      generate all random variables associated with the task, such as random
      starting states, observation noise* etc.

  *If sensor noise is enabled in the MuJoCo model then this will be generated
  using MuJoCo's internal RNG, which has its own independent state.
  """

  def __init__(self, random=None):
    """Initializes a new continuous control task.

    Args:
      random: Optional, either a `numpy.random.RandomState` instance, an integer
        seed for creating a new `RandomState`, or None to select a seed
        automatically (default).
    """
    if not isinstance(random, np.random.RandomState):
      random = np.random.RandomState(random)
    self._random = random
    self._visualize_reward = False

  @property
  def random(self):
    """Task-specific `numpy.random.RandomState` instance."""
    return self._random

  def action_spec(self, physics):
    """Returns a `BoundedArraySpec` matching the `physics` actuators."""
    return mujoco.action_spec(physics)

  def initialize_episode(self, physics):
    """Resets geom colors to their defaults after starting a new episode.

    Subclasses of `base.Task` must delegate to this method after performing
    their own initialization.

    Args:
      physics: An instance of `mujoco.Physics`.
    """
    self.after_step(physics)

  def before_step(self, action, physics):
    """Sets the control signal for the actuators to values in `action`."""
    # Support legacy internal code.
    action = getattr(action, "continuous_actions", action)
    physics.set_control(action)

  def after_step(self, physics):
    """Modifies colors according to the reward."""
    if self._visualize_reward:
      reward = np.clip(self.get_reward(physics), 0.0, 1.0)
      _set_reward_colors(physics, reward)

  @property
  def visualize_reward(self):
    return self._visualize_reward

  @visualize_reward.setter
  def visualize_reward(self, value):
    if not isinstance(value, bool):
      raise ValueError("Expected a boolean, got {}.".format(type(value)))
    self._visualize_reward = value


_MATERIALS = ["self", "effector", "target"]
_DEFAULT = [name + "_default" for name in _MATERIALS]
_HIGHLIGHT = [name + "_highlight" for name in _MATERIALS]


def _set_reward_colors(physics, reward):
  """Sets the highlight, effector and target colors according to the reward."""
  assert 0.0 <= reward <= 1.0
  colors = physics.named.model.mat_rgba
  default = colors[_DEFAULT]
  highlight = colors[_HIGHLIGHT]
  blend_coef = reward ** 4  # Better color distinction near high rewards.
  colors[_MATERIALS] = blend_coef * highlight + (1.0 - blend_coef) * default



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Control Environment tests."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Internal dependencies.

from absl.testing import absltest
from absl.testing import parameterized

from dm_control.rl import control
from dm_env import specs

import mock
import numpy as np

_CONSTANT_REWARD_VALUE = 1.0
_CONSTANT_OBSERVATION = {'observations': np.asarray(_CONSTANT_REWARD_VALUE)}

_ACTION_SPEC = specs.BoundedArray(
    shape=(1,), dtype=np.float, minimum=0.0, maximum=1.0)
_OBSERVATION_SPEC = {'observations': specs.Array(shape=(), dtype=np.float)}


class EnvironmentTest(parameterized.TestCase):

  def setUp(self):
    super(EnvironmentTest, self).setUp()
    self._task = mock.Mock(spec=control.Task)
    self._task.initialize_episode = mock.Mock()
    self._task.get_observation = mock.Mock(return_value=_CONSTANT_OBSERVATION)
    self._task.get_reward = mock.Mock(return_value=_CONSTANT_REWARD_VALUE)
    self._task.get_termination = mock.Mock(return_value=None)
    self._task.action_spec = mock.Mock(return_value=_ACTION_SPEC)
    self._task.observation_spec.side_effect = NotImplementedError()

    self._physics = mock.Mock(spec=control.Physics)
    self._physics.time = mock.Mock(return_value=0.0)

    self._physics.reset_context = mock.MagicMock()

    self._env = control.Environment(physics=self._physics, task=self._task)

  def test_environment_calls(self):
    self._env.action_spec()
    self._task.action_spec.assert_called_with(self._physics)

    self._env.reset()
    self._task.initialize_episode.assert_called_with(self._physics)
    self._task.get_observation.assert_called_with(self._physics)

    action = [1]
    time_step = self._env.step(action)

    self._task.before_step.assert_called()
    self._task.after_step.assert_called_with(self._physics)
    self._task.get_termination.assert_called_with(self._physics)

    self.assertEqual(_CONSTANT_REWARD_VALUE, time_step.reward)

  @parameterized.parameters(
      {'physics_timestep': .01, 'control_timestep': None,
       'expected_steps': 1000},
      {'physics_timestep': .01, 'control_timestep': .05,
       'expected_steps': 5000})
  def test_timeout(self, expected_steps, physics_timestep, control_timestep):
    self._physics.timestep.return_value = physics_timestep
    time_limit = expected_steps * (control_timestep or physics_timestep)
    env = control.Environment(
        physics=self._physics, task=self._task, time_limit=time_limit,
        control_timestep=control_timestep)

    time_step = env.reset()
    steps = 0
    while not time_step.last():
      time_step = env.step([1])
      steps += 1

    self.assertEqual(steps, expected_steps)
    self.assertTrue(time_step.last())

    time_step = env.step([1])
    self.assertTrue(time_step.first())

  def test_observation_spec(self):
    observation_spec = self._env.observation_spec()
    self.assertEqual(_OBSERVATION_SPEC, observation_spec)

  def test_redundant_args_error(self):
    with self.assertRaises(ValueError):
      control.Environment(physics=self._physics, task=self._task,
                          n_sub_steps=2, control_timestep=0.1)

  def test_control_timestep(self):
    self._physics.timestep.return_value = .002
    env = control.Environment(
        physics=self._physics, task=self._task, n_sub_steps=5)
    self.assertEqual(.01, env.control_timestep())

  def test_flatten_observations(self):
    multimodal_obs = dict(_CONSTANT_OBSERVATION)
    multimodal_obs['sensor'] = np.zeros(7, dtype=np.bool)
    self._task.get_observation = mock.Mock(return_value=multimodal_obs)
    env = control.Environment(
        physics=self._physics, task=self._task, flat_observation=True)
    timestep = env.reset()
    self.assertLen(timestep.observation, 1)
    self.assertEqual(timestep.observation[control.FLAT_OBSERVATION_KEY].size,
                     1 + 7)


class ComputeNStepsTest(parameterized.TestCase):

  @parameterized.parameters((0.2, 0.1, 2), (.111, .111, 1), (100, 5, 20),
                            (0.03, 0.005, 6))
  def testComputeNSteps(self, control_timestep, physics_timestep, expected):
    steps = control.compute_n_steps(control_timestep, physics_timestep)
    self.assertEqual(expected, steps)

  @parameterized.parameters((3, 2), (.003, .00101))
  def testComputeNStepsFailures(self, control_timestep, physics_timestep):
    with self.assertRaises(ValueError):
      control.compute_n_steps(control_timestep, physics_timestep)

if __name__ == '__main__':
  absltest.main()



# Copyright 2020 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""A Rodent walker."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import re

from dm_control import composer
from dm_control import mjcf
from dm_control.composer.observation import observable
from dm_control.locomotion.walkers import base
from dm_control.locomotion.walkers import legacy_base
from dm_control.mujoco import wrapper as mj_wrapper
import numpy as np

_XML_PATH = os.path.join(os.path.dirname(__file__),
                         'assets/rodent.xml')

_RAT_MOCAP_JOINTS = [
    'vertebra_1_extend', 'vertebra_2_bend', 'vertebra_3_twist',
    'vertebra_4_extend', 'vertebra_5_bend', 'vertebra_6_twist',
    'hip_L_supinate', 'hip_L_abduct', 'hip_L_extend', 'knee_L', 'ankle_L',
    'toe_L', 'hip_R_supinate', 'hip_R_abduct', 'hip_R_extend', 'knee_R',
    'ankle_R', 'toe_R', 'vertebra_C1_extend', 'vertebra_C1_bend',
    'vertebra_C2_extend', 'vertebra_C2_bend', 'vertebra_C3_extend',
    'vertebra_C3_bend', 'vertebra_C4_extend', 'vertebra_C4_bend',
    'vertebra_C5_extend', 'vertebra_C5_bend', 'vertebra_C6_extend',
    'vertebra_C6_bend', 'vertebra_C7_extend', 'vertebra_C9_bend',
    'vertebra_C11_extend', 'vertebra_C13_bend', 'vertebra_C15_extend',
    'vertebra_C17_bend', 'vertebra_C19_extend', 'vertebra_C21_bend',
    'vertebra_C23_extend', 'vertebra_C25_bend', 'vertebra_C27_extend',
    'vertebra_C29_bend', 'vertebra_cervical_5_extend',
    'vertebra_cervical_4_bend', 'vertebra_cervical_3_twist',
    'vertebra_cervical_2_extend', 'vertebra_cervical_1_bend',
    'vertebra_axis_twist', 'vertebra_atlant_extend', 'atlas', 'mandible',
    'scapula_L_supinate', 'scapula_L_abduct', 'scapula_L_extend', 'shoulder_L',
    'shoulder_sup_L', 'elbow_L', 'wrist_L', 'finger_L', 'scapula_R_supinate',
    'scapula_R_abduct', 'scapula_R_extend', 'shoulder_R', 'shoulder_sup_R',
    'elbow_R', 'wrist_R', 'finger_R'
]


_UPRIGHT_POS = (0.0, 0.0, 0.0)
_UPRIGHT_QUAT = (1., 0., 0., 0.)
_TORQUE_THRESHOLD = 60


class Rat(legacy_base.Walker):
  """A position-controlled rat with control range scaled to [-1, 1]."""

  def _build(self,
             params=None,
             name='walker',
             initializer=None):
    self.params = params
    self._mjcf_root = mjcf.from_path(_XML_PATH)
    if name:
      self._mjcf_root.model = name

    self.body_sites = []
    super(Rat, self)._build(initializer=initializer)

  @property
  def upright_pose(self):
    """Reset pose to upright position."""
    return base.WalkerPose(xpos=_UPRIGHT_POS, xquat=_UPRIGHT_QUAT)

  @property
  def mjcf_model(self):
    """Return the model root."""
    return self._mjcf_root

  @composer.cached_property
  def actuators(self):
    """Return all actuators."""
    return tuple(self._mjcf_root.find_all('actuator'))

  @composer.cached_property
  def root_body(self):
    """Return the body."""
    return self._mjcf_root.find('body', 'torso')

  @composer.cached_property
  def pelvis_body(self):
    """Return the body."""
    return self._mjcf_root.find('body', 'pelvis')

  @composer.cached_property
  def head(self):
    """Return the head."""
    return self._mjcf_root.find('body', 'skull')

  @composer.cached_property
  def left_arm_root(self):
    """Return the left arm."""
    return self._mjcf_root.find('body', 'scapula_L')

  @composer.cached_property
  def right_arm_root(self):
    """Return the right arm."""
    return self._mjcf_root.find('body', 'scapula_R')

  @composer.cached_property
  def ground_contact_geoms(self):
    """Return ground contact geoms."""
    return tuple(
        self._mjcf_root.find('body', 'foot_L').find_all('geom') +
        self._mjcf_root.find('body', 'foot_R').find_all('geom'))

  @composer.cached_property
  def standing_height(self):
    """Return standing height."""
    return self.params['_STAND_HEIGHT']

  @composer.cached_property
  def end_effectors(self):
    """Return end effectors."""
    return (self._mjcf_root.find('body', 'lower_arm_R'),
            self._mjcf_root.find('body', 'lower_arm_L'),
            self._mjcf_root.find('body', 'foot_R'),
            self._mjcf_root.find('body', 'foot_L'))

  @composer.cached_property
  def observable_joints(self):
    """Return observable joints."""
    return tuple(actuator.joint
                 for actuator in self.actuators  #  This lint is mistaken; pylint: disable=not-an-iterable
                 if actuator.joint is not None)

  @composer.cached_property
  def observable_tendons(self):
    return self._mjcf_root.find_all('tendon')

  @composer.cached_property
  def mocap_joints(self):
    return tuple(
        self._mjcf_root.find('joint', name) for name in _RAT_MOCAP_JOINTS)

  @composer.cached_property
  def mocap_joint_order(self):
    return tuple([jnt.name for jnt in self.mocap_joints])  #  This lint is mistaken; pylint: disable=not-an-iterable

  @composer.cached_property
  def bodies(self):
    """Return all bodies."""
    return tuple(self._mjcf_root.find_all('body'))

  @composer.cached_property
  def mocap_bodies(self):
    """Return bodies for mocap comparison."""
    return tuple(body for body in self._mjcf_root.find_all('body')
                 if not re.match(r'(vertebra|hand|toe)', body.name))

  @composer.cached_property
  def primary_joints(self):
    """Return primary (non-vertebra) joints."""
    return tuple(jnt for jnt in self._mjcf_root.find_all('joint')
                 if 'vertebra' not in jnt.name)

  @composer.cached_property
  def vertebra_joints(self):
    """Return vertebra joints."""
    return tuple(jnt for jnt in self._mjcf_root.find_all('joint')
                 if 'vertebra' in jnt.name)

  @composer.cached_property
  def primary_joint_order(self):
    joint_names = self.mocap_joint_order
    primary_names = tuple([jnt.name for jnt in self.primary_joints])  # pylint: disable=not-an-iterable
    primary_order = []
    for nm in primary_names:
      primary_order.append(joint_names.index(nm))
    return primary_order

  @composer.cached_property
  def vertebra_joint_order(self):
    joint_names = self.mocap_joint_order
    vertebra_names = tuple([jnt.name for jnt in self.vertebra_joints])  # pylint: disable=not-an-iterable
    vertebra_order = []
    for nm in vertebra_names:
      vertebra_order.append(joint_names.index(nm))
    return vertebra_order

  @composer.cached_property
  def egocentric_camera(self):
    """Return the egocentric camera."""
    return self._mjcf_root.find('camera', 'egocentric')

  @property
  def _xml_path(self):
    """Return the path to th model .xml file."""
    return self.params['_XML_PATH']

  @composer.cached_property
  def joint_actuators(self):
    """Return all joint actuators."""
    return tuple([act for act in self._mjcf_root.find_all('actuator')
                  if act.joint])

  @composer.cached_property
  def joint_actuators_range(self):
    act_joint_range = []
    for act in self.joint_actuators:  #  This lint is mistaken; pylint: disable=not-an-iterable
      associated_joint = self._mjcf_root.find('joint', act.name)
      act_range = associated_joint.dclass.joint.range
      act_joint_range.append(act_range)
    return act_joint_range

  def pose_to_actuation(self, pose):
    # holds for joint actuators, find desired torque = 0
    # u_ref = [2 q_ref - (r_low + r_up) ]/(r_up - r_low)
    r_lower = np.array([ajr[0] for ajr in self.joint_actuators_range])  #  This lint is mistaken; pylint: disable=not-an-iterable
    r_upper = np.array([ajr[1] for ajr in self.joint_actuators_range])  #  This lint is mistaken; pylint: disable=not-an-iterable
    num_tendon_actuators = len(self.actuators) - len(self.joint_actuators)
    tendon_actions = np.zeros(num_tendon_actuators)
    return np.hstack([tendon_actions, (2*pose[self.joint_actuator_order]-
                                       (r_lower+r_upper))/(r_upper-r_lower)])

  @composer.cached_property
  def joint_actuator_order(self):
    joint_names = self.mocap_joint_order
    joint_actuator_names = tuple([act.name for act in self.joint_actuators])  #  This lint is mistaken; pylint: disable=not-an-iterable
    actuator_order = []
    for nm in joint_actuator_names:
      actuator_order.append(joint_names.index(nm))
    return actuator_order

  def _build_observables(self):
    return RodentObservables(self)


class RodentObservables(legacy_base.WalkerObservables):
  """Observables for the Rat."""

  @composer.observable
  def head_height(self):
    """Observe the head height."""
    return observable.MJCFFeature('xpos', self._entity.head)[2]

  @composer.observable
  def sensors_torque(self):
    """Observe the torque sensors."""
    return observable.MJCFFeature(
        'sensordata',
        self._entity.mjcf_model.sensor.torque,
        corruptor=lambda v, random_state: np.tanh(2 * v / _TORQUE_THRESHOLD)
        )

  @composer.observable
  def tendons_pos(self):
    return observable.MJCFFeature('length', self._entity.observable_tendons)

  @composer.observable
  def tendons_vel(self):
    return observable.MJCFFeature('velocity', self._entity.observable_tendons)

  @composer.observable
  def actuator_activation(self):
    """Observe the actuator activation."""
    model = self._entity.mjcf_model
    return observable.MJCFFeature('act', model.find_all('actuator'))

  @composer.observable
  def appendages_pos(self):
    """Equivalent to `end_effectors_pos` with head's position appended."""

    def relative_pos_in_egocentric_frame(physics):
      end_effectors_with_head = (
          self._entity.end_effectors + (self._entity.head,))
      end_effector = physics.bind(end_effectors_with_head).xpos
      torso = physics.bind(self._entity.root_body).xpos
      xmat = \
          np.reshape(physics.bind(self._entity.root_body).xmat, (3, 3))
      return np.reshape(np.dot(end_effector - torso, xmat), -1)

    return observable.Generic(relative_pos_in_egocentric_frame)

  @property
  def proprioception(self):
    """Return proprioceptive information."""
    return [
        self.joints_pos, self.joints_vel,
        self.tendons_pos, self.tendons_vel,
        self.actuator_activation,
        self.body_height, self.end_effectors_pos, self.appendages_pos,
        self.world_zaxis
    ] + self._collect_from_attachments('proprioception')

  @composer.observable
  def egocentric_camera(self):
    """Observable of the egocentric camera."""

    if not hasattr(self, '_scene_options'):
      # Don't render this walker's geoms.
      self._scene_options = mj_wrapper.MjvOption()
      collision_geom_group = 2
      self._scene_options.geomgroup[collision_geom_group] = 0
      cosmetic_geom_group = 1
      self._scene_options.geomgroup[cosmetic_geom_group] = 0

    return observable.MJCFCamera(self._entity.egocentric_camera,
                                 width=64, height=64,
                                 scene_option=self._scene_options
                                )



# Copyright 2017 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Assets used for testing the MuJoCo bindings."""

import os

from dm_control.utils import io as resources

_ASSETS_DIR = os.path.dirname(__file__)


def get_contents(filename):
  """Returns the contents of an asset as a string."""
  return resources.GetResource(os.path.join(_ASSETS_DIR, filename))


def get_path(filename):
  """Returns the path to an asset."""
  return resources.GetResourceFilename(os.path.join(_ASSETS_DIR, filename))



# Copyright 2020 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Produces reference environments for rodent tasks."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools

from dm_control import composer
from dm_control.composer.variation import distributions
from dm_control.locomotion.arenas import bowl
from dm_control.locomotion.arenas import corridors as corr_arenas
from dm_control.locomotion.arenas import floors
from dm_control.locomotion.arenas import labmaze_textures
from dm_control.locomotion.arenas import mazes
from dm_control.locomotion.props import target_sphere
from dm_control.locomotion.tasks import corridors as corr_tasks
from dm_control.locomotion.tasks import escape
from dm_control.locomotion.tasks import random_goal_maze
from dm_control.locomotion.tasks import reach
from dm_control.locomotion.walkers import rodent

_CONTROL_TIMESTEP = .02
_PHYSICS_TIMESTEP = 0.001


def rodent_escape_bowl(random_state=None):
  """Requires a rodent to climb out of a bowl-shaped terrain."""

  # Build a position-controlled rodent walker.
  walker = rodent.Rat(
      observable_options={'egocentric_camera': dict(enabled=True)})

  # Build a bowl-shaped arena.
  arena = bowl.Bowl(
      size=(20., 20.),
      aesthetic='outdoor_natural')

  # Build a task that rewards the agent for being far from the origin.
  task = escape.Escape(
      walker=walker,
      arena=arena,
      physics_timestep=_PHYSICS_TIMESTEP,
      control_timestep=_CONTROL_TIMESTEP)

  return composer.Environment(time_limit=20,
                              task=task,
                              random_state=random_state,
                              strip_singleton_obs_buffer_dim=True)


def rodent_run_gaps(random_state=None):
  """Requires a rodent to run down a corridor with gaps."""

  # Build a position-controlled rodent walker.
  walker = rodent.Rat(
      observable_options={'egocentric_camera': dict(enabled=True)})

  # Build a corridor-shaped arena with gaps, where the sizes of the gaps and
  # platforms are uniformly randomized.
  arena = corr_arenas.GapsCorridor(
      platform_length=distributions.Uniform(.4, .8),
      gap_length=distributions.Uniform(.05, .2),
      corridor_width=2,
      corridor_length=40,
      aesthetic='outdoor_natural')

  # Build a task that rewards the agent for running down the corridor at a
  # specific velocity.
  task = corr_tasks.RunThroughCorridor(
      walker=walker,
      arena=arena,
      walker_spawn_position=(5, 0, 0),
      walker_spawn_rotation=0,
      target_velocity=1.0,
      contact_termination=False,
      terminate_at_height=-0.3,
      physics_timestep=_PHYSICS_TIMESTEP,
      control_timestep=_CONTROL_TIMESTEP)

  return composer.Environment(time_limit=30,
                              task=task,
                              random_state=random_state,
                              strip_singleton_obs_buffer_dim=True)


def rodent_maze_forage(random_state=None):
  """Requires a rodent to find all items in a maze."""

  # Build a position-controlled rodent walker.
  walker = rodent.Rat(
      observable_options={'egocentric_camera': dict(enabled=True)})

  # Build a maze with rooms and targets.
  wall_textures = labmaze_textures.WallTextures(style='style_01')
  arena = mazes.RandomMazeWithTargets(
      x_cells=11,
      y_cells=11,
      xy_scale=.5,
      z_height=.3,
      max_rooms=4,
      room_min_size=4,
      room_max_size=5,
      spawns_per_room=1,
      targets_per_room=3,
      wall_textures=wall_textures,
      aesthetic='outdoor_natural')

  # Build a task that rewards the agent for obtaining targets.
  task = random_goal_maze.ManyGoalsMaze(
      walker=walker,
      maze_arena=arena,
      target_builder=functools.partial(
          target_sphere.TargetSphere,
          radius=0.05,
          height_above_ground=.125,
          rgb1=(0, 0, 0.4),
          rgb2=(0, 0, 0.7)),
      target_reward_scale=50.,
      contact_termination=False,
      physics_timestep=_PHYSICS_TIMESTEP,
      control_timestep=_CONTROL_TIMESTEP)

  return composer.Environment(time_limit=30,
                              task=task,
                              random_state=random_state,
                              strip_singleton_obs_buffer_dim=True)


def rodent_two_touch(random_state=None):
  """Requires a rodent to tap an orb, wait an interval, and tap it again."""

  # Build a position-controlled rodent walker.
  walker = rodent.Rat(
      observable_options={'egocentric_camera': dict(enabled=True)})

  # Build an open floor arena
  arena = floors.Floor(
      size=(10., 10.),
      aesthetic='outdoor_natural')

  # Build a task that rewards the walker for touching/reaching orbs with a
  # specific time interval between touches
  task = reach.TwoTouch(
      walker=walker,
      arena=arena,
      target_builders=[
          functools.partial(target_sphere.TargetSphereTwoTouch, radius=0.025),
      ],
      randomize_spawn_rotation=True,
      target_type_rewards=[25.],
      shuffle_target_builders=False,
      target_area=(1.5, 1.5),
      physics_timestep=_PHYSICS_TIMESTEP,
      control_timestep=_CONTROL_TIMESTEP,
  )

  return composer.Environment(time_limit=30,
                              task=task,
                              random_state=random_state,
                              strip_singleton_obs_buffer_dim=True)



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================

"""Classes representing observables."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc
import functools

from dm_env import specs
import numpy as np
import six


def _make_aggregator(np_reducer_func, bounds_preserving):
  result = functools.partial(np_reducer_func, axis=0)
  setattr(result, 'bounds_reserving', bounds_preserving)
  return result


AGGREGATORS = {
    'min': _make_aggregator(np.min, True),
    'max': _make_aggregator(np.max, True),
    'mean': _make_aggregator(np.mean, True),
    'median': _make_aggregator(np.median, True),
    'sum': _make_aggregator(np.sum, False),
}


def _get_aggregator(name_or_callable):
  """Returns aggregator from predefined set by name, else returns callable."""
  if name_or_callable is None:
    return None
  elif not callable(name_or_callable):
    try:
      return AGGREGATORS[name_or_callable]
    except KeyError:
      raise KeyError('Unrecognized aggregator name: {!r}. Valid names: {}.'
                     .format(name_or_callable, AGGREGATORS.keys()))
  else:
    return name_or_callable


@six.add_metaclass(abc.ABCMeta)
class Observable(object):
  """Abstract base class for an observable."""

  def __init__(self, update_interval, buffer_size, delay,
               aggregator, corruptor):
    self._update_interval = update_interval
    self._buffer_size = buffer_size
    self._delay = delay
    self._aggregator = _get_aggregator(aggregator)
    self._corruptor = corruptor
    self._enabled = False

  @property
  def update_interval(self):
    return self._update_interval

  @update_interval.setter
  def update_interval(self, value):
    self._update_interval = value

  @property
  def buffer_size(self):
    return self._buffer_size

  @buffer_size.setter
  def buffer_size(self, value):
    self._buffer_size = value

  @property
  def delay(self):
    return self._delay

  @delay.setter
  def delay(self, value):
    self._delay = value

  @property
  def aggregator(self):
    return self._aggregator

  @aggregator.setter
  def aggregator(self, value):
    self._aggregator = _get_aggregator(value)

  @property
  def corruptor(self):
    return self._corruptor

  @corruptor.setter
  def corruptor(self, value):
    self._corruptor = value

  @property
  def enabled(self):
    return self._enabled

  @enabled.setter
  def enabled(self, value):
    self._enabled = value

  @property
  def array_spec(self):
    """The `ArraySpec` which describes observation arrays from this observable.

      If this property is `None`, then the specification should be inferred by
      actually retrieving an observation from this observable.
    """
    return None

  @abc.abstractmethod
  def _callable(self, physics):
    pass

  def observation_callable(self, physics, random_state=None):
    """A callable which returns a (potentially corrupted) observation."""
    raw_callable = self._callable(physics)
    if self._corruptor:
      def _corrupted():
        return self._corruptor(raw_callable(), random_state=random_state)
      return _corrupted
    else:
      return raw_callable

  def __call__(self, physics, random_state=None):
    """Convenience function to just call an observable."""
    return self.observation_callable(physics, random_state)()

  def configure(self, **kwargs):
    """Sets multiple attributes of this observable.

    Args:
      **kwargs: The keyword argument names correspond to the attributes
        being modified.
    Raises:
      AttributeError: If kwargs contained an attribute not in the observable.
    """
    for key, value in six.iteritems(kwargs):
      if not hasattr(self, key):
        raise AttributeError('Cannot add attribute %s in configure.' % key)
      self.__setattr__(key, value)


class Generic(Observable):
  """A generic observable defined via a callable."""

  def __init__(self, raw_observation_callable, update_interval=1,
               buffer_size=None, delay=None,
               aggregator=None, corruptor=None):
    """Initializes this observable.

    Args:
      raw_observation_callable: A callable which accepts a single argument of
        type `control.base.Physics` and returns the observation value.
      update_interval: (optional) An integer, number of simulation steps between
        successive updates to the value of this observable.
      buffer_size: (optional) The maximum size of the returned buffer.
        This option is only relevant when used in conjunction with an
        `observation.Updater`. If None, `observation.DEFAULT_BUFFER_SIZE` will
        be used.
      delay: (optional) Number of additional simulation steps that must be
        taken before an observation is returned. This option is only relevant
        when used in conjunction with an`observation.Updater`. If None,
        `observation.DEFAULT_DELAY` will be used.
      aggregator: (optional) Name of an item in `AGGREGATORS` or a callable that
        performs a reduction operation over the first dimension of the buffered
        observation before it is returned. A value of `None` means that no
        aggregation will be performed and the whole buffer will be returned.
      corruptor: (optional) A callable which takes a single observation as
        an argument, modifies it, and returns it. An example use case for this
        is to add random noise to the observation. When used in a
        `BufferedWrapper`, the corruptor is applied to the observation before
        it is added to the buffer. In particular, this means that the aggregator
        operates on corrupted observations.
    """
    self._raw_callable = raw_observation_callable
    super(Generic, self).__init__(
        update_interval, buffer_size, delay, aggregator, corruptor)

  def _callable(self, physics):
    return lambda: self._raw_callable(physics)


class MujocoFeature(Observable):
  """An observable corresponding to a named MuJoCo feature."""

  def __init__(self, kind, feature_name, update_interval=1,
               buffer_size=None, delay=None,
               aggregator=None, corruptor=None):
    """Initializes this observable.

    Args:
      kind: A string corresponding to a field name in MuJoCo's mjData struct.
      feature_name: A string, or list of strings, or a callable returning
        either, corresponding to the name(s) of an entity in the
        MuJoCo XML model.
      update_interval: (optional) An integer, number of simulation steps between
        successive updates to the value of this observable.
      buffer_size: (optional) The maximum size of the returned buffer.
        This option is only relevant when used in conjunction with an
        `observation.Updater`. If None, `observation.DEFAULT_BUFFER_SIZE` will
        be used.
      delay: (optional) Number of additional simulation steps that must be
        taken before an observation is returned. This option is only relevant
        when used in conjunction with an`observation.Updater`. If None,
        `observation.DEFAULT_DELAY` will be used.
      aggregator: (optional) Name of an item in `AGGREGATORS` or a callable that
        performs a reduction operation over the first dimension of the buffered
        observation before it is returned. A value of `None` means that no
        aggregation will be performed and the whole buffer will be returned.
      corruptor: (optional) A callable which takes a single observation as
        an argument, modifies it, and returns it. An example use case for this
        is to add random noise to the observation. When used in a
        `BufferedWrapper`, the corruptor is applied to the observation before
        it is added to the buffer. In particular, this means that the aggregator
        operates on corrupted observations.
    """
    self._kind = kind
    self._feature_name = feature_name
    super(MujocoFeature, self).__init__(
        update_interval, buffer_size, delay, aggregator, corruptor)

  def _callable(self, physics):
    named_indexer_for_kind = physics.named.data.__getattribute__(self._kind)
    if callable(self._feature_name):
      return lambda: named_indexer_for_kind[self._feature_name()]
    else:
      return lambda: named_indexer_for_kind[self._feature_name]


class MujocoCamera(Observable):
  """An observable corresponding to a MuJoCo camera."""

  def __init__(self, camera_name, height=240, width=320, update_interval=1,
               buffer_size=None, delay=None,
               aggregator=None, corruptor=None, depth=False):
    """Initializes this observable.

    Args:
      camera_name: A string corresponding to the name of a camera in the
        MuJoCo XML model.
      height: (optional) An integer, the height of the rendered image.
      width: (optional) An integer, the width of the rendered image.
      update_interval: (optional) An integer, number of simulation steps between
        successive updates to the value of this observable.
      buffer_size: (optional) The maximum size of the returned buffer.
        This option is only relevant when used in conjunction with an
        `observation.Updater`. If None, `observation.DEFAULT_BUFFER_SIZE` will
        be used.
      delay: (optional) Number of additional simulation steps that must be
        taken before an observation is returned. This option is only relevant
        when used in conjunction with an`observation.Updater`. If None,
        `observation.DEFAULT_DELAY` will be used.
      aggregator: (optional) Name of an item in `AGGREGATORS` or a callable that
        performs a reduction operation over the first dimension of the buffered
        observation before it is returned. A value of `None` means that no
        aggregation will be performed and the whole buffer will be returned.
      corruptor: (optional) A callable which takes a single observation as
        an argument, modifies it, and returns it. An example use case for this
        is to add random noise to the observation. When used in a
        `BufferedWrapper`, the corruptor is applied to the observation before
        it is added to the buffer. In particular, this means that the aggregator
        operates on corrupted observations.
      depth: (optional) A boolean. If `True`, renders a depth image (1-channel)
        instead of RGB (3-channel).
    """
    self._camera_name = camera_name
    self._height = height
    self._width = width

    self._n_channels = 1 if depth else 3
    self._dtype = np.float32 if depth else np.uint8
    self._depth = depth
    super(MujocoCamera, self).__init__(
        update_interval, buffer_size, delay, aggregator, corruptor)

  @property
  def height(self):
    return self._height

  @height.setter
  def height(self, value):
    self._height = value

  @property
  def width(self):
    return self._width

  @width.setter
  def width(self, value):
    self._width = value

  @property
  def array_spec(self):
    return specs.Array(
        shape=(self._height, self._width, self._n_channels), dtype=self._dtype)

  def _callable(self, physics):
    return lambda: physics.render(  # pylint: disable=g-long-lambda
        self._height, self._width, self._camera_name, depth=self._depth)



# Copyright 2018 The dm_control Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Tests for the GLFW based windowing system."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import contextlib
# Internal dependencies.
from absl.testing import absltest
from dm_control.viewer import user_input
import mock
import numpy as np


_OPEN_GL_MOCK = mock.MagicMock()
_GLFW_MOCK = mock.MagicMock()
_MOCKED_MODULES = {
    'OpenGL': _OPEN_GL_MOCK,
    'OpenGL.GL': _OPEN_GL_MOCK,
    'glfw': _GLFW_MOCK,
}
with mock.patch.dict('sys.modules', _MOCKED_MODULES):
  from dm_control.viewer.gui import glfw_gui  # pylint: disable=g-import-not-at-top

glfw_gui.base.GL = _OPEN_GL_MOCK
glfw_gui.base.shaders = _OPEN_GL_MOCK
glfw_gui.glfw = _GLFW_MOCK

# Pretend we are using GLFW for offscreen rendering so that the runtime backend
# check will succeed.
glfw_gui._render.BACKEND = 'glfw'

_EPSILON = 1e-7


class GlfwKeyboardTest(absltest.TestCase):

  def setUp(self):
    super(GlfwKeyboardTest, self).setUp()
    _GLFW_MOCK.reset_mock()
    self.context = mock.MagicMock()
    self.handler = glfw_gui.GlfwKeyboard(self.context)

    self.events = [
        (1, user_input.KEY_T, 't', user_input.PRESS, user_input.MOD_ALT),
        (1, user_input.KEY_T, 't', user_input.RELEASE, user_input.MOD_ALT),
        (1, user_input.KEY_A, 't', user_input.PRESS, 0)]

    self.listener = mock.MagicMock()
    self.handler.on_key += [self.listener]

  def test_single_event(self):
    self.handler._handle_key_event(*self.events[0])
    self.handler.process_events()
    self.listener.assert_called_once_with(user_input.KEY_T, user_input.PRESS,
                                          user_input.MOD_ALT)

  def test_sequence_of_events(self):
    for event in self.events:
      self.handler._handle_key_event(*event)
    self.handler.process_events()
    self.assertEqual(3, self.listener.call_count)
    for event, call_args in zip(self.events, self.listener.call_args_list):
      expected_event = tuple([event[1]] + list(event[-2:]))
      self.assertEqual(expected_event, call_args[0])


class FakePassthroughRenderingContext(object):

  def __init__(self):
    self.window = 0

  def call(self, func, *args):
    return func(*args)


class GlfwMouseTest(absltest.TestCase):

  @contextlib.contextmanager
  def fake_make_current(self):
    yield FakePassthroughRenderingContext()

  def setUp(self):
    super(GlfwMouseTest, self).setUp()
    _GLFW_MOCK.reset_mock()
    _GLFW_MOCK.get_framebuffer_size = mock.MagicMock(return_value=(256, 256))
    _GLFW_MOCK.get_window_size = mock.MagicMock(return_value=(256, 256))
    self.window = mock.MagicMock()
    self.window.make_current = mock.MagicMock(
        side_effect=self.fake_make_current)
    self.handler = glfw_gui.GlfwMouse(self.window)

  def test_moving_mouse(self):
    def move_handler(position, translation):
      self.position = position
      self.translation = translation
    self.new_position = [100, 100]
    self.handler._last_mouse_pos = np.array([99, 101], np.int)
    self.handler.on_move += move_handler

    self.handler._handle_move(self.window, self.new_position[0],
                              self.new_position[1])
    self.handler.process_events()
    np.testing.assert_array_equal(self.new_position, self.position)
    np.testing.assert_array_equal([1, -1], self.translation)

  def test_button_click(self):
    def click_handler(button, action, modifiers):
      self.button = button
      self.action = action
      self.modifiers = modifiers
    self.handler.on_click += click_handler

    self.handler._handle_button(self.window, user_input.MOUSE_BUTTON_LEFT,
                                user_input.PRESS, user_input.MOD_SHIFT)
    self.handler.process_events()

    self.assertEqual(user_input.MOUSE_BUTTON_LEFT, self.button)
    self.assertEqual(user_input.PRESS, self.action)
    self.assertEqual(user_input.MOD_SHIFT, self.modifiers)

  def test_scroll(self):
    def scroll_handler(position):
      self.position = position
    self.handler.on_scroll += scroll_handler

    x_value = 10
    y_value = 20
    self.handler._handle_scroll(self.window, x_value, y_value)
    self.handler.process_events()
    # x_value gets ignored, it's the y_value - the vertical scroll - we're
    # interested in.
    self.assertEqual(y_value, self.position)


class GlfwWindowTest(absltest.TestCase):

  WIDTH = 10
  HEIGHT = 20

  @contextlib.contextmanager
  def fake_make_current(self):
    yield FakePassthroughRenderingContext()

  def setUp(self):
    super(GlfwWindowTest, self).setUp()
    _GLFW_MOCK.reset_mock()
    _GLFW_MOCK.get_video_mode.return_value = (None, None, 60)
    _GLFW_MOCK.get_framebuffer_size.return_value = (4, 5)
    _GLFW_MOCK.get_window_size.return_value = (self.WIDTH, self.HEIGHT)
    self.context = mock.MagicMock()
    self.context.make_current = mock.MagicMock(
        side_effect=self.fake_make_current)
    self.window = glfw_gui.GlfwWindow(
        self.WIDTH, self.HEIGHT, 'title', self.context)

  def test_window_shape(self):
    expected_shape = (self.WIDTH, self.HEIGHT)
    _GLFW_MOCK.get_framebuffer_size.return_value = expected_shape
    self.assertEqual(expected_shape, self.window.shape)

  def test_window_position(self):
    expected_position = (1, 2)
    _GLFW_MOCK.get_window_pos.return_value = expected_position
    self.assertEqual(expected_position, self.window.position)

  def test_freeing_context(self):
    self.window.close = mock.MagicMock()
    self.window.free()
    self.window.close.assert_called_once()

  def test_close(self):
    self.window.close()
    self.context.free.assert_called_once()
    self.assertIsNone(self.window._context)

  def test_closing_window_that_has_already_been_closed(self):
    self.window._context = None
    self.window.close()
    self.assertEqual(0, _GLFW_MOCK.destroy_window.call_count)

  def test_file_drop(self):
    self.expected_paths = ['path1', 'path2']
    def callback(paths):
      self.assertEqual(self.expected_paths, paths)

    was_called_mock = mock.MagicMock()
    self.window.on_files_drop += [callback, was_called_mock]
    self.window._handle_file_drop('window_handle', self.expected_paths)
    was_called_mock.assert_called_once()

  def test_setting_title(self):
    new_title = 'new_title'
    self.window.set_title(new_title)
    self.assertEqual(new_title, _GLFW_MOCK.set_window_title.call_args[0][1])

  def test_enabling_full_screen(self):
    full_screen_pos = (0, 0)
    full_screen_size = (1, 2)
    window_size = (3, 4)
    window_pos = (5, 6)
    reserved_value = 7
    full_size_mode = 8

    _GLFW_MOCK.get_framebuffer_size.return_value = window_size
    _GLFW_MOCK.get_window_pos.return_value = window_pos
    _GLFW_MOCK.get_video_mode.return_value = (
        full_screen_size, reserved_value, full_size_mode)

    self.window.set_full_screen(True)
    _GLFW_MOCK.set_window_monitor.assert_called_once()

    new_position = _GLFW_MOCK.set_window_monitor.call_args[0][2:4]
    new_size = _GLFW_MOCK.set_window_monitor.call_args[0][4:6]
    new_mode = _GLFW_MOCK.set_window_monitor.call_args[0][6]
    self.assertEqual(full_screen_pos, new_position)
    self.assertEqual(full_screen_size, new_size)
    self.assertEqual(full_size_mode, new_mode)


if __name__ == '__main__':
  absltest.main()
